#!/usr/bin/env python
# -*- coding: UTF-8 -*-

# Copyright (c) 2023, Sandflow Consulting LLC
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Data model filter"""

from __future__ import annotations
from typing import Optional

from ttconv.model import ContentDocument
from ttconv.config import ModuleConfiguration

class DocumentFilter:
  """Abstract base class for content document filters"""

  _all_filters = dict()

  def __init__(self, config: ModuleConfiguration) -> None:
    self.config = config

  def process(self, doc: ContentDocument):
    """Processes the specified document in place."""
    raise NotImplementedError

  @classmethod
  def get_config_class(cls) -> ModuleConfiguration:
    """Returns the configuration class for the filter."""
    raise NotImplementedError

  @classmethod
  def get_filter_by_name(cls, name) -> Optional[DocumentFilter]:
    """Returns a list of all document filters"""
    return DocumentFilter._all_filters.get(name)

  def __init_subclass__(cls):
    DocumentFilter._all_filters[cls.get_config_class().name()] = cls

from ttconv.filters.doc import *
